def transform(inventory):
    """Transform an Ansible inventory object prior to JSON serialization.

    This transform is the identity function, and is provided as an example.

    Args:
        inventory: A dictionary containing an Ansible inventory.
            The inventory contains per-host variables for 'ip_address',
            'hostname' and 'mac_address'.

    Returns:
        A dictionary containing an Ansible inventory.
    """
    # # Example: Create two new groups called raw-raspberries and taken-raspberries.
    # #          The first group contains all hosts which which are called
    # #          'raspberrypi', the second group contains the remaining Raspberry Pi
    # #          computers.
    #
    # all_raspberry_hosts = set(inventory['raspberries']['hosts'])
    # raw_raspberry_hosts = set(host for host, hostvars in inventory['_meta']['hostvars'].items()
    #                           if hostvars['hostname'].startswith('raspberrypi'))
    # taken_raspberry_hosts = all_raspberry_hosts - raw_raspberry_hosts
    # inventory['raw-raspberries'] = {'hosts': sorted(raw_raspberry_hosts)}
    # inventory['taken-raspberries'] = {'hosts': sorted(taken_raspberry_hosts)}
    return inventory



import sys
import socket
import ipaddress
import subprocess


_local_networks = ("127.", "0:0:0:0:0:0:0:1")

# ignore these prefixes -- localhost, unspecified, and link-local
_ignored_networks = _local_networks + ("0.", "0:0:0:0:0:0:0:0", "169.254.", "fe80:")


def detect_family(addr):
    if "." in addr:
        assert ":" not in addr
        return socket.AF_INET
    elif ":" in addr:
        return socket.AF_INET6
    else:
        raise ValueError("invalid ipv4/6 address: %r" % addr)


def expand_addr(addr):
    """convert address into canonical expanded form --
    no leading zeroes in groups, and for ipv6: lowercase hex, no collapsed groups.
    """
    family = detect_family(addr)
    addr = socket.inet_ntop(family, socket.inet_pton(family, addr))
    if "::" in addr:
        count = 8-addr.count(":")
        addr = addr.replace("::", (":0" * count) + ":")
        if addr.startswith(":"):
            addr = "0" + addr
    return addr


def _get_local_addr(family, remote):
    try:
        s = socket.socket(family, socket.SOCK_DGRAM)
        try:
            s.connect((remote, 9))
            return s.getsockname()[0]
        finally:
            s.close()
    except socket.error:
        return None


def get_local_addr(remote=None, ipv6=True):
    """Get primary LAN address of host

    Args:
        remote: If supplied, return LAN address that host would use
            to access that specific remote address. By default,
            returns address it would use to access the public
            internet.

        ipv6: Attempt to find an ipv6 address first if True,
            otherwise only check ipv4.

    Returns:
        Primary LAN address for host, or None if could not be
        determined.
    """
    if remote:
        family = detect_family(remote)
        local = _get_local_addr(family, remote)
        if not local:
            return None
        if family == socket.AF_INET6:
            # expand zero groups so the startswith() test works.
            local = expand_addr(local)
        if local.startswith(_local_networks):
            # border case where remote addr belongs to host
            return local
    else:
        # NOTE: the two addresses used here are TESTNET addresses,
        #       which should never exist in the real world.
        if ipv6:
            local = _get_local_addr(socket.AF_INET6, "2001:db8::1234")
            # expand zero groups so the startswith() test works.
            if local:
                local = expand_addr(local)
        else:
            local = None
        if not local:
            local = _get_local_addr(socket.AF_INET, "192.0.2.123")
            if not local:
                return None
    if local.startswith(_ignored_networks):
        return None
    return local


def local_ip_address(remote=None, ipv6=True):
    addr = get_local_addr(remote, ipv6)
    if addr is None:
        return None
    return ipaddress.ip_address(addr)


def local_netmask(local_ip_address):
    a = subprocess.check_output("ifconfig", shell=True).decode(sys.getdefaultencoding())
    matches = filter(lambda line: line.startswith("inet " + str(local_ip_address)),
                   map(str.strip, a.splitlines()))
    try:
        match = next(matches)
    except StopIteration:
        raise RuntimeError("Could not locate network corresponding to {}".format(local_ip_address))

    _, _, suffix = match.partition('netmask')
    netmask_hex = suffix.split()[0]
    netmask_int = int(netmask_hex, base=0)
    netmask_address = ipaddress.ip_address(netmask_int)
    return str(netmask_address)


def make_network(ip_address, netmask):
    ip_address = ipaddress.ip_address(ip_address)
    netmask = ipaddress.ip_address(netmask)
    network_address = ipaddress.ip_address(int(ip_address) & int(netmask))
    return ipaddress.ip_network('/'.join((str(network_address), str(netmask))))


def local_network(remote=None, ipv6=True):
    ip_address = local_ip_address(remote, ipv6)
    netmask = local_netmask(ip_address)
    return make_network(ip_address, netmask)


def arp_ip_mac():
    a = subprocess.check_output("arp -an", shell=True).decode(sys.getdefaultencoding())
    for line in a.splitlines():
        words = line.split()
        maybe_ip_address = [word[1:-1] for word in words if word.count('.') == 3]
        maybe_mac_address = [word for word in words if word.count(':') == 5]
        if maybe_ip_address and maybe_mac_address:
            yield maybe_ip_address[0], maybe_mac_address[0]


def establish_routes():
    """Establish routes to all devices on the local network using nmap.

    This is useful so that subsequent calls to arp_ip_mac() discover all devices.
    """
    network = local_network()
    nmap_command = 'nmap -sP {!s}'.format(network)
    subprocess.check_call(nmap_command, shell=True, stdout=subprocess.DEVNULL)


#!/usr/bin/env python3
"""Pinventory.

An Ansible dynamic inventory provider that finds Raspberry Pi devices.

Usage:
    pinventory --list [--pretty]
    pinventory --host <hostname> [--pretty]
    pinventory --help
    pinventory --version

Options:
    -h --help  Show this screen.
    --version  Show version.
    --list     Show a JSON object of all groups.
    --host     Show a JSON object of variables for the specified host.
    --pretty   Pretty-print JSON output.

"""
import json
import socket
import sys
from functools import reduce

import pkg_resources
from docopt import docopt

from pinventory import __version__
from pinventory.network import arp_ip_mac, establish_routes

RASPBERRY_PI_MAC_ADDRESS_PREFIX = 'b8:27:eb:'
DEFAULT_JSON_INDENTATION = 4

def locate_pi_devices():
    """Find all Raspberry Pi devices on the local network.

    Raspberry Pi computers are identified by the first three
    bytes of their MAC address.
    """
    establish_routes()
    pi_ip_hostname_macs = ip_hostname_macs_matching_mac_prefix(RASPBERRY_PI_MAC_ADDRESS_PREFIX)
    return pi_ip_hostname_macs


def ip_hostname_macs_matching_mac_prefix(mac_prefix):
    pi_ip_mac_addresses = [(ip, mac) for ip, mac in arp_ip_mac() if mac.startswith(mac_prefix)]
    ip_hostname_macs = []
    for pi_ip_address, pi_mac_address in pi_ip_mac_addresses:
        hostaddr, _, _, = socket.gethostbyaddr(pi_ip_address)
        hostname, _, _ = hostaddr.partition('.')
        ip_hostname_macs.append((pi_ip_address, hostname, pi_mac_address))
    return ip_hostname_macs


def make_inventory():
    ip_hostname_macs = locate_pi_devices()
    all_raspberry_group = set(ip for ip, hostname, mac in ip_hostname_macs)
    hostsvars = make_hostsvars(ip_hostname_macs)
    inventory = {
        'raspberries': {
            'hosts': sorted(all_raspberry_group),
        },
        '_meta': {
            'hostvars': hostsvars
        }
    }

    group = 'pinventory.transform.inventory'
    transformers = (entry_point.load() for entry_point in pkg_resources.iter_entry_points(group=group))
    return reduce(lambda a, f: f(a), transformers, inventory)


def make_hostsvars(ip_hostname_macs):
    hostsvars = {ip: {'ip': ip,
                      'hostname': hostname,
                      'mac': mac} for ip, hostname, mac in ip_hostname_macs}

    group = 'pinventory.transform.hostsvars'
    transformers = (entry_point.load() for entry_point in pkg_resources.iter_entry_points(group=group))
    return reduce(lambda a, f: f(a), transformers, hostsvars)


def collect_host_variables(hostname):
    ip_hostname_macs = locate_pi_devices()
    hostsvars = make_hostsvars(ip_hostname_macs)
    try:
        hostvars = hostsvars[hostname]
    except KeyError:
        hostvars = {}
    return hostvars


def print_inventory(indent=None):
    inventory = make_inventory()
    json_inventory = json.dumps(inventory, indent=indent, sort_keys=True)
    print(json_inventory)


def print_host_variables(hostname, indent=None):
    host_variables = collect_host_variables(hostname)
    json_host_variables = json.dumps(host_variables, indent=indent, sort_keys=True)
    print(json_host_variables)


def main(argv=None):
    arguments = docopt(__doc__, argv, version=__version__)
    indent = DEFAULT_JSON_INDENTATION if arguments['--pretty'] else None
    if arguments['--list']:
        print_inventory(indent=indent)
    elif arguments['--host']:
        print_host_variables(arguments['<hostname>'], indent=indent)

if __name__ == '__main__':
    sys.exit(main())





"""A setuptools based setup module for Pinventory.
"""
import os
import sys

MINIMUM_VERSION = (3, 4, 0)

if sys.version_info < MINIMUM_VERSION:
    sys.stderr.write("At least Python {} is required for this package. You are running with Python {}\n"
          .format(
            '.'.join(map(str, MINIMUM_VERSION)),
            '.'.join(map(str, sys.version_info))))
    sys.exit(os.EX_CONFIG)

import io
import re

from codecs import open

from setuptools import setup, find_packages


def read(*names, **kwargs):
    with io.open(
        os.path.join(os.path.dirname(__file__), *names),
        encoding=kwargs.get("encoding", "utf8")
    ) as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

here = os.path.abspath(os.path.dirname(__file__))


# Get the long description from the README file
with open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='pinventory',

    version=find_version("pinventory/__init__.py"),

    description='An Ansible dynamic inventory for locating and categorising Raspberry Pis on the LAN.',
    long_description=long_description,

    # The project's main homepage.
    url='https://github.com/sixty-north/pinventory',

    # Author details
    author='Sixty North AS',
    author_email='rob@sixty-north.com',

    # Choose your license
    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',

        'License :: Other/Proprietary License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.,
        'Programming Language :: Python :: 3.6',
    ],

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['docs', 'test*']),

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        'docopt'
    ],

    # The order of these is important!
    tests_require=[
    ],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
    },

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    package_data={
    },

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    #data_files=[('my_data', ['data/data_file'])],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'console_scripts': [
            'pinventory=pinventory.inventory:main',
        ],
        'pinventory.transform.inventory': [
            'transform=pinventory.transform.inventory:transform'
        ],
        'pinventory.transform.hostsvars': [
            'transform=pinventory.transform.hostsvars:transform'
        ]
    },
)


def transform(hostsvars):
    """Transform an Ansible hostsvars object prior to JSON serialization.

    This transform is the identify function and is provided as an example.

    Args:
        hostsvars: A dictionary mapping hosts (usually IP addresses) to
            dictionaries of variables for each host.

    Returns:
        A modified hostsvars dictionary.
    """
    # # Example: Add a host id variable for each host based on the
    # #          last six digits of its MAC address.
    # for hostvars in hostsvars.values():
    #     hostvars['hostid'] = hostvars['mac'].replace(':', '')[6:]
    return hostsvars
